"""Base class of tokenizer"""
from abc import ABC, abstractmethod
from typing import List, Tuple


class Tokenizer(ABC):
    """
    Abstract class for tokenizing
    ...


    Methods
    -------
    word_tokenize(text: str):
        return tokenized text (abstract method)
    sentence_tokenize(text: str):
        return sentence tokenized text (abstract method)
    """

    @abstractmethod
    def word_tokenize(self, text) -> List[str]:
        """
        Return a tokenized text.
        :param text: the input text
        :return: list of words
        """

    @abstractmethod
    def word_span_tokenize(self, text) -> List[Tuple[int, int, str]]:
        """
        Return spans of tokenized text.
        :param text: the input text
        :return: list of token spans in text
        """

    @abstractmethod
    def sentence_tokenize(self, text) -> List[str]:
        """
        Return a sentence tokenized text.
        :param text: the input text
        :return: list of sentences
        """

    @abstractmethod
    def sentence_span_tokenize(self, text) -> List[Tuple[int, int, str]]:
        """
        Return spans of tokenized text.
        :param text: the input text
        :return: list of token spans in text
        """



"""This module includes Normalizer and NormalizerBuilder"""
from .multi_lingual_normalizer import MultiLingualNormalizer
from .multi_lingual_normalizer_builder import MultiLingualNormalizerBuilder
from .normalizer import Normalizer
from .normalizer_builder import NormalizerBuilder
from .nltk_tokenizer import NltkTokenizer

__all__ = ["Normalizer", "NormalizerBuilder", "NltkTokenizer",
           "MultiLingualNormalizer", "MultiLingualNormalizerBuilder"]



"""This module includes Tokenizer class for tokenizing texts"""
from typing import List, Tuple

import nltk
from nltk.tokenize import TreebankWordTokenizer
from nltk.tokenize.punkt import PunktSentenceTokenizer

from .mappings import MappingDict
from .tokenizer import Tokenizer


class NltkTokenizer(Tokenizer):
    """
    A class for nltk tokenizing.
    ...
    Methods
    -------
    word_tokenize(text: str):
        return tokenized text
    sentence_tokenize(text: str):
        return sentence tokenized text
    """

    def __init__(self, ):
        """
        constructor
        """
        try:
            nltk.data.find('tokenizers/punkt')
        except LookupError:
            print("downloading tokenizer data : ")
            nltk.download('punkt')
        self.__en_mapping = MappingDict.load_jsons(["digit_en", "punc_en"])
        self.__tokenizer = TreebankWordTokenizer()
        self.__sentence_tokenize = PunktSentenceTokenizer()

    def word_tokenize(self, text) -> List[str]:
        tokens = self.word_span_tokenize(text)
        return [text for (_, _, text) in tokens]

    def word_span_tokenize(self, text) -> List[Tuple[int, int, str]]:
        text2 = ''.join(
            [char if not self.__en_mapping.get(char)
             else self.__en_mapping.get(char).char for char in text])
        spans = self.__tokenizer.span_tokenize(text2)
        return [(span[0], span[1], text[span[0]:span[1]]) for span in spans]

    def sentence_tokenize(self, text) -> List[str]:
        tokens = self.sentence_span_tokenize(text)
        return [text for (_, _, text) in tokens]

    def sentence_span_tokenize(self, text) -> List[Tuple[int, int, str]]:
        text2 = ''.join(
            [char if not self.__en_mapping.get(char)
             else self.__en_mapping.get(char).char for char in text])
        spans = self.__sentence_tokenize.span_tokenize(text2)
        return [(span[0], span[1], text[span[0]:span[1]]) for span in spans]



"""This module is for config set """
from __future__ import annotations

from .multi_lingual_normalizer import MultiLingualNormalizer, TokenizationLevel
from .normalizer import Normalizer
from .tokenizer import Tokenizer


class MultiLingualNormalizerBuilder:
    """
    A builder class for multilingual normalizer
    ...

    Attributes
    ----------
    __configs : List[str] or List[Config]
        list of desired configs

    Methods
    -------
    build():
        build an instance of normalizer with current config
    """

    def __init__(self, configs=None):
        """
        constructor
        :param configs default normalizers for each language
        """
        self.__tokenizer = None
        if configs is None:
            configs = {}
        self.__configs = configs
        self.__tokenization_level = TokenizationLevel.SENTENCE_LEVEL
        self.__main_normalizer_lang = 'fa'

    def build(self) -> MultiLingualNormalizer:
        """
        build function of multilingual normalizer
        """
        return MultiLingualNormalizer(self.__configs, self.__main_normalizer_lang,
                                      self.__tokenization_level, self.__tokenizer)

    def tokenization_level(self, tokenization_level: TokenizationLevel) -> MultiLingualNormalizerBuilder:
        """
        Helper function for tokenization level
        """
        self.__tokenization_level = tokenization_level
        return self

    def word_level(self) -> MultiLingualNormalizerBuilder:
        """
        Helper function for tokenization level
        """
        self.__tokenization_level = TokenizationLevel.WORD_LEVEL
        return self

    def sentence_level(self) -> MultiLingualNormalizerBuilder:
        """
        Helper function for tokenization level
        """
        self.__tokenization_level = TokenizationLevel.SENTENCE_LEVEL
        return self

    def lingua_mode(self) -> MultiLingualNormalizerBuilder:
        """
        Helper function for tokenization level
        """
        self.__tokenization_level = TokenizationLevel.LINGUA_MIXED_MODE
        return self

    def main_normalizer_lang(self, main_normalizer_lang: str) -> MultiLingualNormalizerBuilder:
        """
        Helper function for main normalizer language
        """
        self.__main_normalizer_lang = main_normalizer_lang
        return self

    def set_normalizer(self, lang: str, normalizer: Normalizer) -> MultiLingualNormalizerBuilder:
        """
        Set a normalizer for a language
        """
        self.__configs[lang] = normalizer
        return self

    def set_tokenizer(self, tokenizer: Tokenizer) -> MultiLingualNormalizerBuilder:
        """
        Helper function for tokenizer
        """
        self.__tokenizer = tokenizer
        return self



"""This module includes MultiLingualNormalizer class for normalizing texts"""
from enum import Enum
from typing import List, Tuple

from lingua import Language, LanguageDetectorBuilder

from .nltk_tokenizer import NltkTokenizer, Tokenizer
from .normalizer import Normalizer
from .normalizer_builder import NormalizerBuilder


class TokenizationLevel(Enum):
    """
    Level of text for language detection
    """
    WORD_LEVEL = 1
    SENTENCE_LEVEL = 2
    LINGUA_MIXED_MODE = 3


# pylint: disable=too-few-public-methods
class MultiLingualNormalizer:
    """
    Detect lang of text, and then normalize based detected lang
    """

    def __init__(self, configs: dict[str, Normalizer] = None,
                 main_normalizer_lang: str = 'fa',
                 tokenization_level: TokenizationLevel = TokenizationLevel.WORD_LEVEL,
                 tokenizer: Tokenizer = None):
        """ff
        constructor
        :param  configs : List[str]
            list of desired configs
        :param  main_normalizer_lang : str
            that determines main normalization language (for space characters)
        :param  tokenization_level: TokenizationLevel
            tokenization level: word-level, sentence-level and lingua libaray mode
        :param  tokenizer: Tokenizer
            default tokenizer
        """
        # Create a blank Tokenizer with just the English vocab
        if configs is None:
            configs = {
                'fa': NormalizerBuilder().alphabet_fa().alphabet_en().digit_fa().punctuation_fa().diacritic_delete()
                .space_normal().tokenizing().remove_extra_spaces().build(),
                'ar': NormalizerBuilder().alphabet_ar().alphabet_en().digit_ar().punctuation_ar().diacritic_delete()
                .space_normal().tokenizing().remove_extra_spaces().build(),
                'en': NormalizerBuilder().alphabet_en().digit_en().punctuation_en().diacritic_delete()
                .space_normal().tokenizing().remove_extra_spaces().build(),
            }
        self.__configs = configs
        self.__tokenization_level = tokenization_level
        self.__main_normalizer_lang = main_normalizer_lang
        if tokenizer:
            self.__tokenizer = tokenizer
        else:
            self.__tokenizer = NltkTokenizer()

        languages = [Language.PERSIAN, Language.ARABIC, Language.ENGLISH]
        self.__detector = LanguageDetectorBuilder.from_languages(*languages).build()

    def normalize(self, text: str) -> str:
        """
        returns a normalized text
        :param text: the input text, paragraphs are not retained.
        :return: normalized text
        """
        result = ''
        main_normalizer = self.__configs[self.__main_normalizer_lang]
        if self.__tokenization_level is TokenizationLevel.LINGUA_MIXED_MODE:
            lingua = self.__detector.detect_multiple_languages_of(text)
            i = 0
            for det_res in lingua:
                if det_res.start_index > i:
                    result += main_normalizer.normalize(text[i: det_res.start_index])
                normalized = self.__normalize_sub_text(text[det_res.start_index: det_res.end_index], det_res.language)
                result += normalized
                i = det_res.end_index
            if i < len(text):
                result += main_normalizer.normalize(text[i:])
        else:
            spans: List[Tuple[int, int, str]]
            if self.__tokenization_level is TokenizationLevel.WORD_LEVEL:
                spans = self.__tokenizer.word_span_tokenize(text)
            else:
                spans = self.__tokenizer.sentence_span_tokenize(text)
            i = 0
            for (start, end, word) in spans:
                if start > i:
                    result += main_normalizer.normalize(text[i: start])
                normalized = self.__normalize_sub_text(word)
                result += normalized
                i = end
            if i < len(text):
                result += main_normalizer.normalize(text[i:])
        return result

    def __normalize_sub_text(self, sub_text: str, lang: Language | None = None) -> str:
        """
        Normalize text
        :param sub_text:
        :param lang:
        :return:
        """
        language = lang if lang is not None else self.__detector.detect_language_of(sub_text)
        normalizer: Normalizer = self.__configs[self.__main_normalizer_lang]
        match language:
            case Language.ARABIC:
                normalizer = self.__configs['ar']
            case Language.PERSIAN:
                normalizer = self.__configs['fa']
            case Language.ENGLISH:
                normalizer = self.__configs['en']
        return normalizer.normalize(sub_text)



"""This module includes MappingDict class"""
import json
import os
import string
import typing

from typing import Dict
from typing import List

from .char_config import CharConfig


class MappingDict:
    """
    A class for mapping funcs.

    ...

    Attributes
    ----------
    Methods
    -------
    """

    @staticmethod
    def load_jsons(configs: List[str]) -> Dict[str, CharConfig]:
        """
            Mapping for configs and english configs from files
            :param configs: the input configs
            :return: mapping configs and mapping english
        """
        all_configs = []
        current_directory = os.path.dirname(os.path.abspath(__file__))
        for dir_path, _, filenames in os.walk(current_directory + "/data/"):
            for filename in filenames:
                abspath = os.path.abspath(os.path.join(dir_path, filename))
                all_configs.extend(MappingDict.read_json(abspath))

        return MappingDict.get_mapping(configs, all_configs)

    @staticmethod
    def get_mapping(configs: List[str],
                    all_configs: List[Dict[str, typing.Any]]) -> Dict[str, CharConfig]:
        """
            get mapping of desired configs
            :param configs: input configs
            :param all_configs: the all configs extracted from files
            :return: mapping char to char
        """
        mapping = {}
        if configs and len(configs) == 0:
            return mapping
        for data in all_configs:
            for key in data["map"].keys():
                if key in configs:
                    key_map = data["map"][key]
                    mapping[key_map["char"]] = CharConfig.from_dict(data, key)
                    for char_dic in data["others"]:
                        char = char_dic["char"]
                        if not mapping.get(char):
                            mapping[char] = CharConfig.from_dict(data, key)
        return mapping

    @staticmethod
    def read_json(address: string):
        """
            return loaded json from files
            :param address: the input address
            :return: loaded json
        """
        with open(address, encoding="utf-8") as json_file:
            data = json.load(json_file)
            return data



# testing Fibonacci number function
# pylint: skip-file
from ..src import NormalizerBuilder, MultiLingualNormalizerBuilder

fa_normalizer = NormalizerBuilder().alphabet_fa().alphabet_en().digit_fa().punctuation_fa() \
    .diacritic_delete().space_normal().tokenizing().remove_extra_spaces().build()
ar_normalizer = NormalizerBuilder().alphabet_ar().alphabet_en().digit_ar().punctuation_ar() \
    .diacritic_delete().space_normal().tokenizing().remove_extra_spaces().build()


def test_object():
    norm = MultiLingualNormalizerBuilder().build()
    assert norm is not None


def test_multi_lingual_normalization():
    text = "در این آیه کمی تفکر کنید.\nالَّذِينَ يُقِيمُونَ الصَّلاةَ وَيُؤْتُونَ الزَّكَاةَ"
    target = "در این آیه کمی تفکر کنید.\nالذین یقیمون الصلاة ویؤتون الزکاة"
    norm = MultiLingualNormalizerBuilder().set_normalizer('fa', fa_normalizer)\
        .set_normalizer('ar', ar_normalizer).main_normalizer_lang('fa').build()
    assert target == norm.normalize(text)

    text = "در این آیه کمی تفکر کنید: الَّذِينَ يُقِيمُونَ الصَّلاةَ وَيُؤْتُونَ الزَّكَاةَ"
    target = "در این آیه کمی تفکر کنید: الذین یقیمون الصلات ویوتون الزکات"
    norm = MultiLingualNormalizerBuilder().set_normalizer('fa', fa_normalizer) \
        .set_normalizer('ar', ar_normalizer).main_normalizer_lang('fa')\
        .sentence_level().build()
    assert target == norm.normalize(text)

    text = "در این آیه کمی تفکر کنید: الَّذِينَ يُقِيمُونَ الصَّلاةَ وَيُؤْتُونَ الزَّكَاةَ"
    target = "در این آیه کمی تفکر کنید: الذین یقیمون الصلاة ویؤتون الزکاة"
    norm = MultiLingualNormalizerBuilder().set_normalizer('fa', fa_normalizer) \
        .set_normalizer('ar', ar_normalizer).main_normalizer_lang('fa')\
        .word_level().build()
    assert target == norm.normalize(text)

    text = "در این 1 آیه کمی تفکر کنید: الَّذِينَ يُقِيمُونَ الصَّلاةَ وَيُؤْتُونَ الزَّكَاةَ"
    target = "در این ۱ آیه کمی تفکر کنید: الذین یقیمون الصلاة ویؤتون الزکاة"
    norm = MultiLingualNormalizerBuilder().set_normalizer('fa', fa_normalizer) \
        .set_normalizer('ar', ar_normalizer).main_normalizer_lang('fa').word_level().build()
    assert target == norm.normalize(text)